/*------------------------------------------------------------------------
 |  FILE            Connection.java
 |  PART OF         Sverre's Simple JDBC Driver for PostgreSQL
 |  WRITTEN BY      Sverre H. Huseby <shh@thathost.com>
 +----------------------------------------------------------------------*/

package no.shhsoft.postgresql;

import java.util.*;
import java.net.*;
import java.io.*;
import java.sql.*;

/**
 * A Connection represents a session with a specific database. Within
 * the context of a Connection, SQL statements are executed and
 * results are returned.
 * <P>
 * Connection's database is able to provide information describing its
 * tables, its supported SQL grammar, its stored procedures, the
 * capabilities of this connection, etc. This information is obtained
 * with the getMetaData method.
 * <P>
 * <B>Note:</B> By default the Connection automatically commits
 * changes after executing each statement. If auto commit has been
 * disabled, an explicit commit must be done or database changes will
 * not be saved.
 *
 * @see      DriverManager#getConnection
 * @see      Statement
 * @see      ResultSet
 * @see      DatabaseMetaData
 * @author   Sverre H. Huseby
 *           &lt;<A HREF="mailto:shh@thathost.com">shh@thathost.com</A>&gt;
 * @version $Id: Connection.java,v 1.10 2002/02/05 22:54:25 sverrehu Exp $ */
public final class Connection
implements java.sql.Connection {
/*-----------------------------------------------------------------------+
|  PRIVATE PART                                                          |
+-----------------------------------------------------------------------*/
    private static int nextCursorNumber = 0;

    private PostgreSQLCommunicator pg;
    private boolean doAutoCommit = true;

    private Vector statements;
    private Hashtable oidToSQLTypeHash;  /* Integer, Integer */
    private Hashtable oidToTypeNameHash; /* Integer, String */

    private DatabaseMetaData metaData;

    private Object    grabLock = new Object();
    private Statement grabber = null;
    private Thread    grabbingThread = null;

    private int transactionIsolationLevel
        = Connection.TRANSACTION_READ_COMMITTED;

    /* TODO: do something useful with the warnings */
    private SQLWarning firstWarning = null;


    private synchronized void addStatement(Statement stmt) {
        statements.addElement(stmt);
    }

    private synchronized void closeStatements()
    throws SQLException {
        int q;

        for (q = statements.size() - 1; q >= 0; q--)
            ((Statement) statements.elementAt(q)).close();
    }

    private synchronized void closeResultSets()
    throws SQLException {
        int q;

        for (q = statements.size() - 1; q >= 0; q--)
            ((Statement) statements.elementAt(q)).setCurrentResultSet(null);
    }

    private void readTypeCache()
    throws SQLException {
        Statement stmt;
        ResultSet rs;
        int       oid, sqlType;
        String    typeName;
        Integer   i;

        oidToSQLTypeHash = new Hashtable();
        oidToTypeNameHash = new Hashtable();
        stmt = (Statement) createStatement();
        rs = (ResultSet) stmt.executeQuery("SELECT oid,typname FROM pg_type");
        while (rs.next()) {
            oid = rs.getInt(1);
            typeName = rs.getString(2);
            if (typeName.compareTo("int2") == 0)
                sqlType = Types.SMALLINT;
            else if (typeName.compareTo("int4") == 0
                     || typeName.compareTo("oid") == 0)
                sqlType = Types.INTEGER;
            else if (typeName.compareTo("int8") == 0)
                sqlType = Types.BIGINT;
            else if (typeName.compareTo("money") == 0)
                sqlType = Types.DECIMAL;
            else if (typeName.compareTo("numeric") == 0)
                sqlType = Types.NUMERIC;
            else if (typeName.compareTo("float4") == 0)
                sqlType = Types.REAL;
            else if (typeName.compareTo("float8") == 0)
                sqlType = Types.DOUBLE;
            else if (typeName.compareTo("bpchar") == 0
                     || typeName.compareTo("char2") == 0
                     || typeName.compareTo("char4") == 0
                     || typeName.compareTo("char8") == 0
                     || typeName.compareTo("char16") == 0)
                sqlType = Types.CHAR;
            else if (typeName.compareTo("varchar") == 0
                     || typeName.compareTo("text") == 0
                     || typeName.compareTo("name") == 0
                     || typeName.compareTo("filename") == 0)
                sqlType = Types.VARCHAR;
            else if (typeName.compareTo("bool") == 0)
                sqlType = Types.BIT;
            else if (typeName.compareTo("date") == 0)
                sqlType = Types.DATE;
            else if (typeName.compareTo("time") == 0)
                sqlType = Types.TIME;
            else if (typeName.compareTo("abstime") == 0
                     || typeName.compareTo("timestamp") == 0)
                sqlType = Types.TIMESTAMP;
            else
                sqlType = Types.OTHER;
            i = new Integer(oid);
            oidToSQLTypeHash.put(i, new Integer(sqlType));
            oidToTypeNameHash.put(i, typeName);
        }
        rs.close();
        stmt.close();
    }

/*-----------------------------------------------------------------------+
|  PACKAGE LOCAL PART                                                    |
+-----------------------------------------------------------------------*/
    String url, username;  /* used by DatabaseMetaData */

    final synchronized void removeStatement(Statement stmt) {
        statements.removeElement(stmt);
    }

    Connection(String url, String host, int port, String database,
               String username, String password)
    throws SQLException {
        this.url = url;
        this.username = username;

        pg = new PostgreSQLCommunicator(host, port, database,
                                        username, password);

        statements = new Vector();
        readTypeCache();

        metaData = new DatabaseMetaData(this);
    }

    final synchronized int query(String query, ResultSet rs)
    throws SQLException {
        assertConnectionOpen();
        return pg.query(query, rs);
    }

    final synchronized int query(String query)
    throws SQLException {
        assertConnectionOpen();
        return pg.query(query);
    }

    final String getUniqueCursorName() {
        StringBuffer sb;

        /* TODO: somehow make it unique across several VMs */
        sb = new StringBuffer();
        sb.append((char) 'c');
        sb.append(nextCursorNumber++);
        return sb.toString();
    }

    final int getSQLType(int typeOID) {
        Integer sqlType;

        sqlType = (Integer) oidToSQLTypeHash.get(new Integer(typeOID));
        if (sqlType == null)
            return Types.OTHER;
        return sqlType.intValue();
    }

    final String getTypeName(int typeOID) {
        return (String) oidToTypeNameHash.get(new Integer(typeOID));
    }

    final void grabConnection(Statement stmt)
    throws SQLException {
        synchronized (grabLock) {
            if (grabber == stmt) {
                System.err.println("warning: statement tried to grab "
                                   + "connection a second time.");
                return;
            }
            /* if the connection is already grabbed, and this thread
             * is not the same as the grebbing thread, we may get
             * around by blocking this thread until the connection is
             * released.  if, however, the thread that already holds
             * the grab attempts to grab for a second statement, we
             * can not block, as this would (of course) block forever.
             * the only solution in that case is to throw an
             * exception. */
            if (grabber != null && grabbingThread == Thread.currentThread()) {
System.err.println("second block attempt from one thread.  throwing exception.");
System.out.println("thread: " + Thread.currentThread().toString());
(new Throwable()).printStackTrace();
                throw new SQLException(
                  "Attempt to start a new query on a connection when another\n"
                + "query was still active on that connection.\n"
                + "PostgreSQL does not support that.");
            }
            while (grabber != null) {
System.err.println("connection already grabbed.  waiting...");
System.out.println("thread: " + Thread.currentThread().toString());
(new Throwable()).printStackTrace();
                try {
                    grabLock.wait();
                } catch (InterruptedException e) {
                    throw new SQLException(
                         "Timout waiting for our turn on the connection."
                         + "  Another statement active.");
                }
            }
//  System.err.println("grabbing connection");
//  System.out.println("thread: " + Thread.currentThread().toString());
//  (new Throwable()).printStackTrace();
            grabber = stmt;
            grabbingThread = Thread.currentThread();
        }
    }

    final void releaseConnection(Statement stmt)
    throws SQLException {
        synchronized (grabLock) {
            if (grabber != stmt) {
                System.err.println("warning: non-grabbing statement tried "
                                   + "to release connection.");
                return;
            }
//  System.err.println("releasing the connection");
//  System.out.println("thread: " + Thread.currentThread().toString());
//  (new Throwable()).printStackTrace();
            grabber = null;
            grabbingThread = null;
            grabLock.notify();
        }
    }

    final void assertConnectionOpen()
    throws SQLException {
        if (pg.isClosed())
            throw new SQLException("Database connection not open");
    }

    final void abortQuery(ResultSet rs)
    throws SQLException {
        pg.abortQuery(rs);
    }

/*-----------------------------------------------------------------------+
|  INHERITANTS' INTERFACE                                                |
+-----------------------------------------------------------------------*/
    /* java.lang.Object *************************************************/
    protected void finalize() {
        try {
            close();
        } catch (SQLException e) {
        }
    }

/*-----------------------------------------------------------------------+
|  PUBLIC INTERFACE                                                      |
+-----------------------------------------------------------------------*/

    /* java.sql.Connection **********************************************/
    /**
     * SQL statements without parameters are normally executed using
     * Statement objects. If the same SQL statement is executed many
     * times, it is more efficient to use a PreparedStatement
     *
     * @return     a new Statement object
     * @exception  SQLException if a database-access error occurs.
     */
    public java.sql.Statement createStatement()
    throws SQLException {
        Statement stmt;

        stmt = new Statement(this);
        addStatement(stmt);
        return stmt;
        /* TODO ? */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public PreparedStatement prepareStatement(String sql)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public CallableStatement prepareCall(String sql)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * A driver may convert the JDBC sql grammar into its system's
     * native SQL grammar prior to sending it; nativeSQL returns the
     * native form of the statement that the driver would have sent.
     *
     * @param      sql     a SQL statement that may contain one or
     *                     more '?' parameter placeholders
     * @return     the native form of this statement
     * @exception  SQLException if a database-access error occurs.
     */
    public String nativeSQL(String sql)
    throws SQLException {
        return sql;
        /* TODO: parse escape syntax etc. */
    }

    /**
     * If a connection is in auto-commit mode, then all its SQL
     * statements will be executed and committed as individual
     * transactions.  Otherwise, its SQL statements are grouped into
     * transactions that are terminated by either commit() or
     * rollback().  By default, new connections are in auto-commit
     * mode.
     * <P>
     * The commit occurs when the statement completes or the next
     * execute occurs, whichever comes first. In the case of
     * statements returning a ResultSet, the statement completes when
     * the last row of the ResultSet has been retrieved or the
     * ResultSet has been closed. In advanced cases, a single
     * statement may return multiple results as well as output
     * parameter values. Here the commit occurs when all results and
     * output param values have been retrieved.
     *
     * @param      autoCommit
     *                     true enables auto-commit; false disables
     *                     auto-commit.
     * @exception  SQLException if a database-access error occurs.
     */
    public void setAutoCommit(boolean autoCommit)
    throws SQLException {
        if (autoCommit == doAutoCommit)
            return;
        closeResultSets();
        if (autoCommit)
            query("END");
        else
            query("BEGIN");
        doAutoCommit = autoCommit;
        /* TODO ? */
    }

    /**
     * Gets the current auto-commit state.
     *
     * @return     Current state of auto-commit mode.
     * @exception  SQLException if a database-access error occurs.
     * @see        #setAutoCommit
     */
    public boolean getAutoCommit()
    throws SQLException {
        return doAutoCommit;
    }

    /**
     * Makes all changes made since the previous commit/rollback
     * permanent and releases any database locks currently held by the
     * Connection. This method should only be used when auto commit
     * has been disabled.
     *
     * @exception  SQLException if a database-access error occurs.
     * @see        #setAutoCommit
     */
    public synchronized void commit()
    throws SQLException {
        if (doAutoCommit)
            return;

        closeResultSets();
        query("COMMIT");
        query("BEGIN");
        /* TODO ? */
    }

    /**
     * Drops all changes made since the previous commit/rollback and
     * releases any database locks currently held by the
     * Connection. This method should only be used when auto commit
     * has been disabled.
     *
     * @exception  SQLException if a database-access error occurs.
     * @see        #setAutoCommit
     */
    public synchronized void rollback()
    throws SQLException {
        if (doAutoCommit)
            return;

        closeResultSets();
        query("ROLLBACK");
        query("BEGIN");
        /* TODO ? */
    }

    /**
     * In some cases, it is desirable to immediately release a
     * Connection's database and JDBC resources instead of waiting for
     * them to be automatically released; the close method provides
     * this immediate release.
     * <P>
     * <B>Note:</B> A Connection is automatically closed when it is
     * garbage collected. Certain fatal errors also result in a closed
     * Connection.
     *
     * @exception  SQLException if a database-access error occurs.
     */
    public void close()
    throws SQLException {
        closeStatements();
        pg.close();
    }

    /**
     * Tests to see if a Connection is closed.
     *
     * @return     true if the connection is closed; false if it's still open
     * @exception  SQLException if a database-access error occurs.
     */
    public boolean isClosed()
    throws SQLException {
        return pg.isClosed();
    }

    /**
     * A Connection's database is able to provide information
     * describing its tables, its supported SQL grammar, its stored
     * procedures, the capabilities of this connection, etc. This
     * information is made available through a DatabaseMetaData
     * object.
     *
     * @return     a DatabaseMetaData object for this Connection
     * @exception  SQLException if a database-access error occurs.
     */
    public java.sql.DatabaseMetaData getMetaData()
    throws SQLException {
        return metaData;
    }

    /**
     * You can put a connection in read-only mode as a hint to enable
     * database optimizations.
     * <P>
     * <B>Note:</B> setReadOnly cannot be called while in the middle
     * of a transaction.
     *
     * @param      readOnly
     *                     true enables read-only mode; false disables
     *                     read-only mode.
     * @exception  SQLException if a database-access error occurs.
     */
    public void setReadOnly(boolean readOnly)
    throws SQLException {
        /* do nothing, as PostgreSQL doesn't support read only mode. */
    }

    /**
     * Tests to see if the connection is in read-only mode.
     *
     * @return     true if connection is read-only
     * @exception  SQLException if a database-access error occurs.
     */
    public boolean isReadOnly()
    throws SQLException {
        /* we're never in read only mode. */
        return false;
    }

    /**
     * A sub-space of this Connection's database may be selected by
     * setting a catalog name. If the driver does not support catalogs
     * it will silently ignore this request.
     *
     * @exception  SQLException if a database-access error occurs.
     */
    public void setCatalog(String catalog)
    throws SQLException {
        /* catalogs not supported. */
    }

    /**
     * Return the Connection's current catalog name.
     *
     * @return     the current catalog name or null
     * @exception  SQLException if a database-access error occurs.
     */
    public String getCatalog()
    throws SQLException {
        /* catalogs are not supported. */
        return null;
    }

    /**
     * Transactions are not supported.
     */
    public static final int TRANSACTION_NONE = 0;

    /**
     * Dirty reads, non-repeatable reads and phantom reads can occur.
     */
    public static final int TRANSACTION_READ_UNCOMMITTED = 1;

    /**
     * Dirty reads are prevented; non-repeatable reads and phantom
     * reads can occur.
     */
    public static final int TRANSACTION_READ_COMMITTED = 2;

    /**
     * Dirty reads and non-repeatable reads are prevented; phantom
     * reads can occur.
     */
    public static final int TRANSACTION_REPEATABLE_READ = 4;

    /**
     * Dirty reads, non-repeatable reads and phantom reads are
     * prevented.
     */
    public static final int TRANSACTION_SERIALIZABLE = 8;

    /**
     * You can call this method to try to change the transaction
     * isolation level using one of the TRANSACTION_* values.
     * <P>
     * <B>Note:</B> setTransactionIsolation cannot be called while in
     * the middle of a transaction.
     *
     * @param      level   one of the TRANSACTION_* isolation values
     *                     with the exception of TRANSACTION_NONE;
     *                     some databases may not support other values
     * @exception  SQLException if a database-access error occurs.
     * @see        DatabaseMetaData#supportsTransactionIsolationLevel
     */
    public void setTransactionIsolation(int level)
    throws SQLException {
        /* TODO: make sure this isn't called within a transaction,
         * or while a result set is active. */
        if (level == Connection.TRANSACTION_READ_COMMITTED)
            query("SET TRANSACTION ISOLATION LEVEL READ COMMITTED");
        else if (level == Connection.TRANSACTION_SERIALIZABLE)
            query("SET TRANSACTION ISOLATION LEVEL SERIALIZABLE");
        else
            throw new SQLException("Transaction isolation " + level
                                   + " not supported");
        transactionIsolationLevel = level;
    }

    /**
     * Get this Connection's current transaction isolation mode.
     *
     * @return     the current TRANSACTION_* mode value
     * @exception  SQLException if a database-access error occurs.
     */
    public int getTransactionIsolation()
    throws SQLException {
        return transactionIsolationLevel;
    }

    /**
     * The first warning reported by calls on this Connection is
     * returned.
     * <P>
     * <B>Note:</B> Subsequent warnings will be chained to this
     * SQLWarning.
     *
     * @return     the first SQLWarning or null
     * @exception  SQLException if a database-access error occurs.
     */
    public SQLWarning getWarnings()
    throws SQLException {
        return firstWarning;
    }

    /**
     * After this call, getWarnings returns null until a new warning
     * is reported for this Connection.
     *
     * @exception  SQLException if a database-access error occurs.
     */
    public void clearWarnings()
    throws SQLException {
        firstWarning = null;
    }

    /* -- JDBC 2.0 ---------------------------------------------------- */

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.Statement createStatement(int resultSetType,
                                              int resultSetConcurrency)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.PreparedStatement prepareStatement(String sql,
                                                     int resultSetType,
                                                     int resultSetConcurrency)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.CallableStatement prepareCall(String sql,
                                                  int resultSetType,
                                                  int resultSetConcurrency)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.util.Map getTypeMap()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void setTypeMap(java.util.Map map)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }
}
