/*------------------------------------------------------------------------
 |  FILE            CursoredResultSet.java
 |  PART OF         Sverre's Simple JDBC Driver for PostgreSQL
 |  WRITTEN BY      Sverre H. Huseby <shh@thathost.com>
 +----------------------------------------------------------------------*/

package no.shhsoft.postgresql;

import java.util.*;
import java.io.*;
import java.sql.*;
import java.math.*;

/**
 * A ResultSet provides access to a table of data generated by
 * executing a Statement. The table rows are retrieved in
 * sequence. Within a row its column values can be accessed in any
 * order.
 * <P>
 * A ResultSet maintains a cursor pointing to its current row of data.
 * Initially the cursor is positioned before the first row.  The
 * 'next' method moves the cursor to the next row.
 * <P>
 * The getXXX methods retrieve column values for the current row.  You
 * can retrieve values either using the index number of the column, or
 * by using the name of the column.  In general using the column index
 * will be more efficient.  Columns are numbered from 1.
 * <P>
 * For maximum portability, ResultSet columns within each row should
 * be read in left-to-right order and each column should be read only
 * once.
 * <P>
 * For the getXXX methods, the JDBC driver attempts to convert the
 * underlying data to the specified Java type and returns a suitable
 * Java value.  See the JDBC specification for allowable mappings from
 * SQL types to Java types with the ResultSet.getXXX methods.
 * <P>
 * Column names used as input to getXXX methods are case insensitive.
 * When performing a getXXX using a column name, if several columns
 * have the same name, then the value of the first matching column
 * will be returned. The column name option is designed to be used
 * when column names are used in the SQL query. For columns that are
 * NOT explicitly named in the query, it is best to use column
 * numbers. If column names were used there is no way for the
 * programmer to guarantee that they actually refer to the intended
 * columns.
 * <P>
 * A ResultSet is automatically closed by the Statement that generated
 * it when that Statement is closed, re-executed, or is used to
 * retrieve the next result from a sequence of multiple results.
 * <P>
 * The number, types and properties of a ResultSet's columns are
 * provided by the ResulSetMetaData object returned by the getMetaData
 * method.
 *
 * @see      Statement#executeQuery
 * @see      Statement#getResultSet
 * @see      ResultSetMetaData
 * @author   Sverre H. Huseby
 *           &lt;<A HREF="mailto:shh@thathost.com">shh@thathost.com</A>&gt;
 * @version  $Id: CursoredResultSet.java,v 1.8 2002/02/05 22:54:25 sverrehu Exp $
 */
public final class CursoredResultSet
extends ResultSet {
/*-----------------------------------------------------------------------+
|  PRIVATE PART                                                          |
+-----------------------------------------------------------------------*/
    private String fetchCommand;
    private String cursorName;

/*-----------------------------------------------------------------------+
|  PACKAGE LOCAL PART                                                    |
+-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------+
|  INHERITANTS' INTERFACE                                                |
+-----------------------------------------------------------------------*/
    CursoredResultSet(Connection conn, Statement stmt, String cursorName)
    throws SQLException {
        super(conn, stmt);

        this.cursorName = cursorName;
        /* need to create the correct fetch statement, so we have to
         * redo the following call not that the cursor name is known.
         * the call is done from super() too. */
        setFetchSize(DEFAULT_FETCH_SIZE);
    }

    CursoredResultSet()
    throws SQLException {
        this(null, null, null);
    }

    /* the sql string should already have been unescaped, and this
     * result set must be the active result set of the connection. */
    void startFetching(String sql)
    throws SQLException {
        StringBuffer sb;

        sb = new StringBuffer();
        sb.append("DECLARE ");
        sb.append(cursorName);
        sb.append(" CURSOR FOR ");
        sb.append(sql);

        try {
            if (conn.getAutoCommit())
                conn.query("BEGIN");
            conn.query(sb.toString());
        } catch (SQLException e) {
            stmt.stopFetching();
            throw e;
        }
    }

    /* java.lang.Object *************************************************/
    protected void finalize() {
        try {
            close();
        } catch (SQLException e) {
        }
    }

/*-----------------------------------------------------------------------+
|  PUBLIC INTERFACE                                                      |
+-----------------------------------------------------------------------*/
    /* java.sql.ResultSet ***********************************************/
    /**
     * A ResultSet is initially positioned before its first row; the
     * first call to next makes the first row the current row; the
     * second call makes the second row the current row, etc.
     * <P>
     * If an input stream from the previous row is open, it is
     * implicitly closed. The ResultSet's warning chain is cleared
     * when a new row is read.
     *
     * @return     true if the new current row is valid; false if there
     *             are no more rows
     * @exception  SQLException if a database-access error occurs.
     */
    public synchronized boolean next()
    throws SQLException {
        assertOpen();

        if (conn == null || stmt.haveStoppedFetching)
            return false;

        try {
            stmt.clearWarnings();
            synchronized (conn) {
                if (++currRow < numRows)
                    return true;

                numRows = 0;
                currRow = -1;

                try {
                    fetchWithoutResult = false;
                    conn.query(fetchCommand, this);
                } catch (SQLException e) {
                    if (conn.getAutoCommit())
                        conn.query("ROLLBACK");
                    stmt.stopFetching();
                    throw e;
                }

                /* TODO ? */
                if (fetchWithoutResult || numRows == 0) {
                    if (conn.getAutoCommit())
                        conn.query("END");
                    stmt.stopFetching();
                    return false;
                }

                currRow = 0;
                return true;
            }
        } catch (SQLException e) {
            /* make sure connection lock is released. */
            stmt.stopFetching();
            throw e;
        }
    }

    /**
     * In some cases, it is desirable to immediately release a
     * ResultSet's database and JDBC resources instead of waiting for
     * this to happen when it is automatically closed; the close
     * method provides this immediate release.
     * <P>
     * <B>Note:</B> A ResultSet is automatically closed by the
     * Statement that generated it when that Statement is closed,
     * re-executed, or is used to retrieve the next result from a
     * sequence of multiple results. A ResultSet is also automatically
     * closed when it is garbage collected.
     *
     * @exception  SQLException if a database-access error occurs.
     */
    public synchronized void close()
    throws SQLException {
        if (stmt != null && !stmt.haveStoppedFetching && conn.getAutoCommit())
            conn.query("END");
        super.close();
        /* TODO */
    }

    /* java.sql.ResultSet for JDBC 2.0 **********************************/
    /* only methods that make sense due to how this driver is implemented.
     * this is _not_ a JDBC 2.0 driver (yet). */

    public synchronized void setFetchSize(int rows)
    throws SQLException {
        StringBuffer sb;

        super.setFetchSize(rows);

        sb = new StringBuffer();
        sb.append("FETCH ");
        sb.append(rows);
        sb.append(" IN ");
        sb.append(cursorName);
        fetchCommand = sb.toString();
    }
}
