/*------------------------------------------------------------------------
 |  FILE            ResultSetMetaData.java
 |  PART OF         Sverre's Simple JDBC Driver for PostgreSQL
 |  WRITTEN BY      Sverre H. Huseby <shh@thathost.com>
 +----------------------------------------------------------------------*/

package no.shhsoft.postgresql;

import java.sql.*;

/**
 * A ResultSetMetaData object can be used to find out about the types
 * and properties of the columns in a ResultSet.
 *
 * @author   Sverre H. Huseby
 *           &lt;<A HREF="mailto:shh@thathost.com">shh@thathost.com</A>&gt;
 * @version  $Id: ResultSetMetaData.java,v 1.3 2002/02/05 22:54:26 sverrehu Exp $
 */
public final class ResultSetMetaData
implements java.sql.ResultSetMetaData {
/*-----------------------------------------------------------------------+
|  PRIVATE PART                                                          |
+-----------------------------------------------------------------------*/
    private ResultSet  resultSet;
    private Connection conn;

/*-----------------------------------------------------------------------+
|  PACKAGE LOCAL PART                                                    |
+-----------------------------------------------------------------------*/
    ResultSetMetaData(ResultSet resultSet, Connection conn) {
        this.resultSet = resultSet;
        this.conn = conn;
    }

/*-----------------------------------------------------------------------+
|  PUBLIC INTERFACE                                                      |
+-----------------------------------------------------------------------*/

    /* java.sql.ResultSetMetaData ***************************************/
    /**
     * What's the number of columns in the ResultSet?
     *
     * @return     the number
     * @exception  SQLException if a database-access error occurs.
     */
    public int getColumnCount()
    throws SQLException {
        return resultSet.numFields;
    }

    /**
     * Is the column automatically numbered, thus read-only?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     true if so
     * @exception  SQLException if a database-access error occurs.
     */
    public boolean isAutoIncrement(int column)
    throws SQLException {
        return false;  /* not supported in PostgreSQL. */
    }

    /**
     * Does a column's case matter?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     true if so
     * @exception  SQLException if a database-access error occurs.
     */
    public boolean isCaseSensitive(int column)
    throws SQLException {
        int type;

        type = getColumnType(column);
        return !(type == Types.BIT
                 || type == Types.TINYINT
                 || type == Types.SMALLINT
                 || type == Types.INTEGER
                 || type == Types.BIGINT
                 || type == Types.FLOAT
                 || type == Types.REAL
                 || type == Types.DOUBLE
                 || type == Types.NUMERIC
                 || type == Types.DECIMAL
                 || type == Types.DATE
                 || type == Types.TIME
                 || type == Types.TIMESTAMP);
    }

    /**
     * Can the column be used in a where clause?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     true if so
     * @exception  SQLException if a database-access error occurs.
     */
    public boolean isSearchable(int column)
    throws SQLException {
        return true;
        /* TODO: may not be so. */
    }

    /**
     * Is the column a cash value?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     true if so
     * @exception  SQLException if a database-access error occurs.
     */
    public boolean isCurrency(int column)
    throws SQLException {
        String typeName;

        typeName = getColumnTypeName(column);
        /* actually, the money type is documented as obsolete, but
         * we test for it anyway. */
        return (typeName != null && typeName.compareTo("money") == 0);
    }

    /**
     * Can you put a NULL in this column?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     columnNoNulls, columnNullable or columnNullableUnknown
     * @exception  SQLException if a database-access error occurs.
     */
    public int isNullable(int column)
    throws SQLException {
        /* TODO: may ask some system table, perhaps? */
        return ResultSetMetaData.columnNullableUnknown;
    }

    /**
     * Does not allow NULL values.
     */
    public static final int columnNoNulls = 0;

    /**
     * Allows NULL values.
     */
    public static final int columnNullable = 1;

    /**
     * Nullability unknown.
     */
    public static final int columnNullableUnknown = 2;

    /**
     * Is the column a signed number?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     true if so
     * @exception  SQLException if a database-access error occurs.
     */
    public boolean isSigned(int column)
    throws SQLException {
        int type;

        type = getColumnType(column);
        return (type == Types.TINYINT
                || type == Types.SMALLINT
                || type == Types.INTEGER
                || type == Types.BIGINT
                || type == Types.FLOAT
                || type == Types.REAL
                || type == Types.DOUBLE
                || type == Types.NUMERIC
                || type == Types.DECIMAL);
    }

    /**
     * What's the column's normal max width in chars?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     max width
     * @exception  SQLException if a database-access error occurs.
     */
    public int getColumnDisplaySize(int column)
    throws SQLException {
        int type, fieldTypeSize, fieldTypeModifier;

        fieldTypeSize = resultSet.fieldTypeSize[column - 1];
        fieldTypeModifier = resultSet.fieldTypeModifier[column - 1];
        type = getColumnType(column);

        switch (type) {
          case Types.BIT: /* 0 or 1 */
            return 1;
          case Types.TINYINT: /* 0 - 255 */
            return 3;
          case Types.SMALLINT: /* -32768 - 32767 */
            return 6;
          case Types.INTEGER: /* -2147483648 - 2147483647 */
            return 11;
          case Types.BIGINT: /* -9223372036854775808 - 9223372036854775807 */
            return 20;
          case Types.FLOAT:
            return 20;
          case Types.REAL:
            return 11;
          case Types.DOUBLE:
            return 20;
          case Types.NUMERIC: /* user defined precision */
          case Types.DECIMAL: /* user defined precision */
            return ((fieldTypeModifier >> 16) & 0xFFFF)
                + (fieldTypeModifier & 0xFFFF) + 1;
          case Types.CHAR: /* user defined precision */
          case Types.VARCHAR: /* user defined precision */
            return fieldTypeModifier;
          case Types.LONGVARCHAR: /* user defined precision */
            return fieldTypeModifier;  /* actually, we don't use this one. */
          case Types.DATE: /* 2000-02-24 */
            return 10;
          case Types.TIME: /* 23:11:01 */
            return 8;
          case Types.TIMESTAMP: /* 2000-02-24 23:11:01+01 */
            return 22;
          case Types.BINARY: /* user defined precision */
          case Types.VARBINARY: /* user defined precision */
          case Types.LONGVARBINARY: /* user defined precision */
            return fieldTypeModifier;  /* actually, we don't use these. */
          case Types.NULL:
            return 4;
        }
        /* have no idea if we fall out here... */
        return fieldTypeSize;
    }

    /**
     * What's the suggested column title for use in printouts and
     * displays?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     true if so
     * @exception  SQLException if a database-access error occurs.
     */
    public String getColumnLabel(int column)
    throws SQLException {
        return getColumnName(column);
    }

    /**
     * What's a column's name?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     column name
     * @exception  SQLException if a database-access error occurs.
     */
    public String getColumnName(int column)
    throws SQLException {
        return resultSet.fieldName[column - 1];
    }

    /**
     * What's a column's table's schema?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     schema name or "" if not applicable
     * @exception  SQLException if a database-access error occurs.
     */
    public String getSchemaName(int column)
    throws SQLException {
        return "";
    }

    /**
     * What's a column's number of decimal digits?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     precision
     * @exception  SQLException if a database-access error occurs.
     */
    public int getPrecision(int column)
    throws SQLException {
        int type, fieldTypeModifier;

        fieldTypeModifier = resultSet.fieldTypeModifier[column - 1];
        type = getColumnType(column);

        switch (type) {
          case Types.BIT: /* 0 or 1 */
            return 1;
          case Types.TINYINT: /* 0 - 255 */
            return 3;
          case Types.SMALLINT: /* -32768 - 32767 */
            return 5;
          case Types.INTEGER: /* -2147483648 - 2147483647 */
            return 10;
          case Types.BIGINT: /* -9223372036854775808 - 9223372036854775807 */
            return 19;
          case Types.FLOAT:
            return 15;
          case Types.REAL:
            return 7;
          case Types.DOUBLE:
            return 15;
          case Types.NUMERIC: /* user defined precision */
          case Types.DECIMAL: /* user defined precision */
            return ((fieldTypeModifier >> 16) & 0xFFFF);
        }
        return 0;  /* not a number. */
    }

    /**
     * What's a column's number of digits to right of the decimal point?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     scale
     * @exception  SQLException if a database-access error occurs.
     */
    public int getScale(int column)
    throws SQLException {
        int type, fieldTypeModifier;

        fieldTypeModifier = resultSet.fieldTypeModifier[column - 1];
        type = getColumnType(column);

        switch (type) {
          case Types.BIT: /* 0 or 1 */
          case Types.TINYINT: /* 0 - 255 */
          case Types.SMALLINT: /* -32768 - 32767 */
          case Types.INTEGER: /* -2147483648 - 2147483647 */
          case Types.BIGINT: /* -9223372036854775808 - 9223372036854775807 */
            return 0;
          case Types.FLOAT:
            return 15;
          case Types.REAL:
            return 7;
          case Types.DOUBLE:
            return 15;
          case Types.NUMERIC: /* user defined precision */
          case Types.DECIMAL: /* user defined precision */
            return (fieldTypeModifier & 0xFFFF);
        }
        return 0;  /* not a number. */
    }

    /**
     * What's a column's table name?
     *
     * @return     table name or "" if not applicable
     * @exception  SQLException if a database-access error occurs.
     */
    public String getTableName(int column)
    throws SQLException {
        /* TODO: how do I get this?  and how about views, or other
         * queries that do not select from a table? */
        return "";
    }

    /**
     * What's a column's table's catalog name?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     column name or "" if not applicable.
     * @exception  SQLException if a database-access error occurs.
     */
    public String getCatalogName(int column)
    throws SQLException {
        return "";  /* no catalogs. */
    }

    /**
     * What's a column's SQL type?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     SQL type
     * @exception  SQLException if a database-access error occurs.
     * @see        Types
     */
    public int getColumnType(int column)
    throws SQLException {
        return conn.getSQLType(resultSet.fieldTypeOID[column - 1]);
    }

    /**
     * What's a column's data source specific type name?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     type name
     * @exception  SQLException if a database-access error occurs.
     */
    public String getColumnTypeName(int column)
    throws SQLException {
        return conn.getTypeName(resultSet.fieldTypeOID[column - 1]);
    }

    /**
     * Is a column definitely not writable?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     true if so
     * @exception  SQLException if a database-access error occurs.
     */
    public boolean isReadOnly(int column)
    throws SQLException {
        /* we don't support positioned update, so I guess it will
         * always be read only. */
        return true;
    }

    /**
     * Is it possible for a write on the column to succeed?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     true if so
     * @exception  SQLException if a database-access error occurs.
     */
    public boolean isWritable(int column)
    throws SQLException {
        return !isReadOnly(column);
    }

    /**
     * Will a write on the column definitely succeed?
     *
     * @param      column  the first column is 1, the second is 2, ...
     * @return     true if so
     * @exception  SQLException if a database-access error occurs.
     */
    public boolean isDefinitelyWritable(int column)
    throws SQLException {
        return false;  /* who knows what happens to the local power plant? */
    }

    /* -- JDBC 2.0 ---------------------------------------------------- */

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public String getColumnClassName(int column)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }
}
