/*------------------------------------------------------------------------
 |  FILE            Statement.java
 |  PART OF         Sverre's Simple JDBC Driver for PostgreSQL
 |  WRITTEN BY      Sverre H. Huseby <shh@thathost.com>
 +----------------------------------------------------------------------*/

package no.shhsoft.postgresql;

import java.io.*;
import java.sql.*;

/**
 * A Statement object is used for executing a static SQL statement and
 * obtaining the results produced by it.
 * <P>
 * Only one ResultSet per Statement can be open at any point in
 * time. Therefore, if the reading of one ResultSet is interleaved
 * with the reading of another, each must have been generated by
 * different Statements. All statement execute methods implicitly
 * close a statment's current ResultSet if an open one exists.
 *
 * @see      Connection#createStatement
 * @see      ResultSet
 * @author   Sverre H. Huseby
 *           &lt;<A HREF="mailto:shh@thathost.com">shh@thathost.com</A>&gt;
 * @version  $Id: Statement.java,v 1.9 2002/02/05 22:54:26 sverrehu Exp $
 */
public final class Statement
implements java.sql.Statement {
/*-----------------------------------------------------------------------+
|  PRIVATE PART                                                          |
+-----------------------------------------------------------------------*/
    private Connection conn;
    private ResultSet  currentResultSet = null;
    private boolean    doEscapeProcessing = true;
    private String     cursorName = null;
    /* TODO: do something useful about these warnings */
    private SQLWarning firstWarning = null;

    private boolean isCursorableQuery(String sql) {
        return Util.toLowerCase(sql.trim()).startsWith("select");
    }

    private ResultSet createResultSet()
    throws SQLException {
        ResultSet rs;

        rs = new ResultSet();
        setCurrentResultSet(rs);
        return rs;
    }

    private String processEscapes(String sql)
    throws SQLException {
        if (!doEscapeProcessing)
            return sql;
        /* TODO */
        return sql;
    }

/*-----------------------------------------------------------------------+
|  PACKAGE LOCAL PART                                                    |
+-----------------------------------------------------------------------*/
    boolean haveStoppedFetching = true;

    Statement(Connection conn) {
        this.conn = conn;
    }

    void stopFetching()
    throws SQLException {
        if (haveStoppedFetching)
            return;
        haveStoppedFetching = true;
//              conn.setActiveResultSet(null);
        setCurrentResultSet(null);
    }

    synchronized void setCurrentResultSet(ResultSet rs)
    throws SQLException {
        ResultSet tmp;

        if (rs == currentResultSet)
            return;
        if (currentResultSet != null) {
            /* temporary object needed to avoid infinite recursion. */
            tmp = currentResultSet;
            currentResultSet = null;
            tmp.close();
        }
//              conn.setActiveResultSet(rs);
        if (rs == null)
            conn.releaseConnection(this);
        else
            conn.grabConnection(this);
        currentResultSet = rs;
    }

    synchronized void abortCurrentResultSet()
    throws SQLException {
        if (currentResultSet == null)
            return;
        conn.abortQuery(currentResultSet);
        setCurrentResultSet(null);
    }

/*-----------------------------------------------------------------------+
|  INHERITANTS' INTERFACE                                                |
+-----------------------------------------------------------------------*/
    /* java.lang.Object *************************************************/
    protected void finalize() {
        try {
            close();
        } catch (SQLException e) {
        }
    }

/*-----------------------------------------------------------------------+
|  PUBLIC INTERFACE                                                      |
+-----------------------------------------------------------------------*/
    /* java.sql.Statement ***********************************************/
    /**
     * Executes a SQL statement that returns a single ResultSet.
     *
     * @param      sql     typically this is a static SQL SELECT statement
     * @return     a ResultSet that contains the data produced by the
     *             query; never null
     * @exception  SQLException if a database-access error occurs.
     */
    public synchronized java.sql.ResultSet executeQuery(String sql)
    throws SQLException {
        abortCurrentResultSet();
        try {
//  System.err.println("NOTE! using test version without cursor!");
//              if (false) {
            if (isCursorableQuery(sql)) {
                CursoredResultSet rs;
                String            tmpCursorName;

                tmpCursorName = cursorName;
                if (tmpCursorName == null)
                    tmpCursorName = conn.getUniqueCursorName();
                rs = new CursoredResultSet(conn, this, tmpCursorName);
                setCurrentResultSet(rs);
                cursorName = tmpCursorName;
                haveStoppedFetching = false;
                rs.startFetching(processEscapes(sql));

                return rs;
            } else {
                /* TODO: this is not implemented.  we probably do not
                 * need it either. */
                ResultSet rs;

                rs = new ResultSet(conn, this);
                setCurrentResultSet(rs);
                haveStoppedFetching = false;
                /* TODO */

                return rs;
            }
        } catch (SQLException e) {
            abortCurrentResultSet();
            throw e;
        }
    }

    /**
     * Executes a SQL INSERT, UPDATE or DELETE statement. In addition,
     * SQL statements that return nothing such as SQL DDL statements
     * can be executed.
     *
     * @param      sql     a SQL INSERT, UPDATE or DELETE statement
     *                     or a SQL statement that returns nothing
     * @return     either the row count for INSERT, UPDATE or DELETE or 0
     *             for SQL statements that return nothing
     * @exception  SQLException if a database-access error occurs.
     */
    public synchronized int executeUpdate(String sql)
    throws SQLException {
        int ret;

        abortCurrentResultSet();
        conn.grabConnection(this);
        try {
            ret = conn.query(processEscapes(sql));
        } finally {
            conn.releaseConnection(this);
        }
        return ret;
        /* TODO ? */
    }

    /**
     * In many cases, it is desirable to immediately release a
     * Statements's database and JDBC resources instead of waiting for
     * this to happen when it is automatically closed; the close
     * method provides this immediate release.
     * <P>
     * <B>Note:</B> A Statement is automatically closed when it is
     * garbage collected. When a Statement is closed, its current
     * ResultSet, if one exists, is also closed.
     *
     * @exception  SQLException if a database-access error occurs.
     */
    public synchronized void close()
    throws SQLException {
        if (conn != null)
            conn.removeStatement(this);
        stopFetching();
        setCurrentResultSet(null);
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public int getMaxFieldSize()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void setMaxFieldSize(int max)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public int getMaxRows()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void setMaxRows(int max)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * If escape scanning is on (the default), the driver will do
     * escape substitution before sending the SQL to the database.
     * <P>
     * Note: Since prepared statements have usually been parsed prior
     * to making this call, disabling escape processing for prepared
     * statements will like have no affect.
     *
     * @param      enable  true to enable; false to disable
     * @exception  SQLException if a database-access error occurs.
     */
    public void setEscapeProcessing(boolean enable)
    throws SQLException {
        doEscapeProcessing = enable;
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public int getQueryTimeout()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void setQueryTimeout(int seconds)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void cancel()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * The first warning reported by calls on this Statement is
     * returned.  A Statment's execute methods clear its SQLWarning
     * chain. Subsequent Statement warnings will be chained to this
     * SQLWarning.
     * <P>
     * The warning chain is automatically cleared each time a
     * statement is (re)executed.
     * <P>
     * <B>Note:</B> If you are processing a ResultSet then any
     * warnings associated with ResultSet reads will be chained on the
     * ResultSet object.
     *
     * @return     the first SQLWarning or null
     * @exception  SQLException if a database-access error occurs.
     */
    public SQLWarning getWarnings()
    throws SQLException {
        return firstWarning;
    }

    /**
     * After this call, getWarnings returns null until a new warning
     * is reported for this Statement.
     *
     * @exception  SQLException if a database-access error occurs.
     */
    public void clearWarnings()
    throws SQLException {
        firstWarning = null;
    }

    /**
     * Defines the SQL cursor name that will be used by subsequent
     * Statement execute methods. This name can then be used in SQL
     * positioned update/delete statements to identify the current row
     * in the ResultSet generated by this statement.  If the database
     * doesn't support positioned update/delete, this method is a
     * noop.
     * <P>
     * <B>Note:</B> By definition, positioned update/delete execution
     * must be done by a different Statement than the one which
     * generated the ResultSet being used for positioning. Also,
     * cursor names must be unique within a Connection.
     *
     * @param      name   the new cursor name.
     * @exception  SQLException if a database-access error occurs.
     */
    public void setCursorName(String name)
    throws SQLException {
        /* TODO: this can not be done if a result set is currently open. */
        cursorName = name;
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean execute(String sql)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * Returns the current result as a ResultSet.  It should only be
     * called once per result.
     *
     * @return     the current result as a ResultSet; null if the result
     *             is an update count or there are no more results
     * @exception  SQLException if a database-access error occurs.
     * @see        #execute
     */
    public java.sql.ResultSet getResultSet()
    throws SQLException {
        if (haveStoppedFetching)
            return null;
        return currentResultSet;
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public int getUpdateCount()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public boolean getMoreResults()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /* -- JDBC 2.0 ---------------------------------------------------- */

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void setFetchDirection(int direction)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public int getFetchDirection()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void setFetchSize(int rows)
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public int getFetchSize()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public int getResultSetConcurrency()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public int getResultSetType()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void addBatch( String sql )
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public void clearBatch()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public int[] executeBatch()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }

    /**
     * @exception  SQLException since the method is not implemented.
     */
    public java.sql.Connection getConnection()
    throws SQLException {
        throw new NotImplementedException();
        /* TODO */
    }
}
